package gpsys.cos;

import gpsys.*;
import gpsys.primitives.*;

/**
 * The Fitness class for the terrain robot problem.  The fitness is
 * measured using several values.  The first is the number of time spent on the terrain
 * when the evolved program has completed. The second value is
 * the energy left in the robot, number of lamps not maxed.  The fitness of one GP is 
 * deemed better than another if either the error is smaller or if the error is equal, but
 * the complexity is smaller.  The termination criteria is met when the robot is dead,
 * the lamps are all maxed, or no energy is left in the system.
 *
 * @version             1.1, 24 Mars 1999
 * @author      <a href="mailto:epierre@e-nef.com">Emmanuel PIERRE</a>
 *                      <address>DEA IARFA,</address>
 */
public class CosSimFitness extends Fitness {
	/**
	 * The time lived on the terrain.
	 */
	int		fitness;
	/**
	 * The complexity of the individual i.e. the number of Genes it contains.
	 */
	int		complexity;
	/**
	 * The terrain to be cut.
	 */
	Terrain	terrain;
	/*
	 * The robot to be controlled by the program.
	 */
	Robot	robot;

	/**
	 * Construct a Fitness object with default Fitness.
	 */
	public CosSimFitness(Terrain terrain, Robot robot) {
		this.terrain  	= terrain;
		this.robot 	= robot;
		fitness		= 0;
		complexity	= 0;
	}

	/**
	 * Constructs a Fitness object by evaluating the specified individual.
	 * The raw fitness is the time spent on the terrain, the robot alived, and 
	 * the number of lamps not maxed when the GP evolved code has executed.
	 *
	 * @param terrain	The terrain to be proceeded by the robot.
	 * @param robot		The terrain robot to use.
	 * @param gpParameters 	The GP parameters for this run.
	 * @param i		The individual to be evaluated.
	 */
	public CosSimFitness(Terrain terrain, Robot robot,
		GPParameters gpParameters, Individual i) {

		this.terrain  = terrain;
		this.robot = robot;

		terrain.reinit();
		robot.reset();
		int cnt=0; // not to loop forever when the robot s thinking and thus not moving

		try {
		while (((robot.movementAvailable>0)&&((terrain.l0+terrain.l1+terrain.l2)<(terrain.lmax*3)))&&(cnt<300)) {
				i.evaluateObject();
				cnt++;
			}
		}
		catch (GPException e) {
			gpParameters.observer.exception(e);
		}

		// TEST FITNESS 
		fitness = 304-terrain.ttl-((terrain.l0<terrain.lmax)?1:0)-((terrain.l1<terrain.lmax)?1:0)-((terrain.l2<terrain.lmax)?1:0)-((robot.movementAvailable>0)?1:0);

		complexity = i.complexity();

	}

	/**
	 * Adds a fitness value to this fitness value.
	 *
	 * @param f	the fitness value to be added.
	 */
	public void add(Fitness f) {
		fitness += ((CosSimFitness)f).fitness;
		//complexity  += ((CosSimFitness)f).complexity;
	};

	/**
	 * Divide the fitness by the specified integer.  This is used by the
	 * GP system to calculate the average fitness of the population.
	 *
	 * @param	divisor	the integer to divide the fitness by.
	 */
	public void divide(int divisor) {
		fitness /= divisor;
	}

	/**
	 * Tests if this fitness value is greater than another fitness value.
	 *
	 * @param	f	the fitness with which to compare.
	 * @return	true if f has higher fitness, false otherwise.
	 */
	public boolean greaterThan(Fitness f) {
		if (fitness < ((CosSimFitness)f).fitness)
			return true;
		return false;
	}

	/**
	 * Tests if this fitness is less than another fitness.
	 *
	 * @param	f	the fitness with which to compare.
	 * @return	true if f has less fitness, false otherwise.
	 */
	public boolean lessThan(Fitness f) {
		if (fitness > ((CosSimFitness)f).fitness)
			return true;
		return false;
	}

	/**
	 * Tests if this fitness is equal to another fitness.
	 *
	 * @param	f	the fitness with which to compare.
	 * @return	true if f has the same fitness, false otherwise.
	 */
	public boolean equals(Fitness f) {
		return
			(fitness == ((CosSimFitness)f).fitness) ;
	}

	/**
	 * Creates a new instance of the Fitness object with a default fitness.
	 * Note that the same Terrain and Robot are used in the new instance.
	 *
	 * @return	an instance of a Fitness object with default fitness.
	 */
	public Fitness instance() {
		return new CosSimFitness(terrain, robot);
	}

	/**
	 * Creates a new instance of the Fitness object which represents the
	 * fitness of the specified individual.
	 * Note that the same Terrain and Robot are used in the new instance.
	 *
	 * @param	gpParameters	the parameters for this GP run.
	 * @param	i				the individual to be evaluated.
	 * @return	The Fitness of the specified individual.
	 */
	public Fitness instance(GPParameters gpParameters, Individual i) {
		return new CosSimFitness(terrain, robot, gpParameters, i);
	}

	/**
	 * Tests whether this fitness meets the termination criteria.
	 *
	 * @return	true if the termination criteria has been met, false otherwise.
	 */
	public boolean terminationCondition() {
		//return ((fitness == 0)||(robot.movementAvailable==0));
		//return ((fitness == 0)||(robot.movementAvailable==0)||(((terrain.l0+terrain.l1+terrain.l2)==(terrain.lmax*3))));
		return ((fitness == 0)||(((terrain.l0+terrain.l1+terrain.l2)==(terrain.lmax*3))&&(robot.movementAvailable==0)));
	}

	/**
	 * Converts the fitness into a String suitable for printing.
	 *
	 * @return	A String representing the fitness.
	 */
	public String toString() {
		return "Fitness(F1:" + fitness + ",R1:"+ robot.movementAvailable +",T:"+terrain.ttl+",L0:"+terrain.l0+",L1:"+terrain.l1+",L2:"+terrain.l2+",CS:"+terrain.csnrj+")";
	}
}
